# SPDX-FileCopyrightText: PhiBo DinoTools (2021)
# SPDX-License-Identifier: GPL-3.0-or-later

import re
from typing import Dict, List, Set

import click
import nagiosplugin

from ..cli import cli
from ..helper import logger, RouterOSVersion
from ..resource import RouterOSCheckResource


class SystemTemperatureResource(RouterOSCheckResource):
    name = "Temperature"

    def __init__(
        self,
        cmd_options,
        check: nagiosplugin.Check,
        warning_values: List[str],
        critical_values: List[str],
        use_regex: bool
    ):
        super().__init__(cmd_options=cmd_options)

        self._check = check

        self.names: Set[str] = set()
        self.values: Dict[str, float] = {}
        self.use_regex: bool = use_regex

        self.warning_values: Dict[str, str] = {}
        self.critical_values: Dict[str, str] = {}
        self.warning_regex_values: Dict[re.Pattern, str] = {}
        self.critical_regex_values: Dict[re.Pattern, str] = {}

        if self.use_regex:
            self.warning_regex_values = self.prepare_regex_thresholds(warning_values)
            self.critical_regex_values = self.prepare_regex_thresholds(critical_values)
        else:
            self.warning_values = self.prepare_thresholds(warning_values)
            self.critical_values = self.prepare_thresholds(critical_values)

        self._fetch_data()

    def _fetch_data(self):
        logger.info("Fetching data ...")
        call = self.api.path(
            "/system/health"
        )
        api_result_items = tuple(call)
        if self.routeros_version < RouterOSVersion("7"):
            api_result_items = self._convert_v6_list_to_v7(api_result_items)

        regex_name = re.compile(r".*temperature.*")
        for item in api_result_items:
            m = regex_name.match(item["name"])
            if not m:
                continue

            if self.use_regex:
                for regex, threshold in self.warning_regex_values.items():
                    if regex.match(item["name"]):
                        self.warning_values[item["name"]] = threshold
                        break

                for regex, threshold in self.critical_regex_values.items():
                    if regex.match(item["name"]):
                        self.critical_values[item["name"]] = threshold
                        break

            self.names.add(item["name"])
            self.values[item["name"]] = float(item["value"])

    def probe(self):
        for name, value in self.values.items():
            self._check.add(nagiosplugin.ScalarContext(
                name=name,
                warning=self.warning_values.get(name),
                critical=self.critical_values.get(name),
            ))
            yield nagiosplugin.Metric(
                name=name,
                value=value,
            )


@cli.command("system.temperature")
@click.option(
    "warning_values",
    "--value-warning",
    multiple=True,
    help=(
        "Set a warning threshold for a value. "
        "Example: If cpu-temperature should be in the range of 40 and 60°C you can set "
        "--value-warning cpu-temperature:40:60 "
        "If cpu-temperature should not be higher than 50.5°C you can set "
        "--value-warning cpu-temperature:50.5 "
        "Can be specified multiple times"
    )
)
@click.option(
    "critical_values",
    "--value-critical",
    multiple=True,
    help=(
        "Set a critical threshold for a value. "
        "Example: If cpu-temperature should be in the range of 40 and 60°C you can set "
        "--value-critical cpu-temperature:40:60 "
        "If cpu-temperature should not be higher than 50.5°C you can set "
        "--value-critical cpu-temperature:50.5 "
        "Can be specified multiple times"
    )
)
@click.option(
    "--regex",
    "use_regex",
    default=False,
    is_flag=True,
    help="Treat values from --value-warning and --value-critical as regex to find all matching values"
)
@click.option(
    "--no-temperature-ok",
    is_flag=True,
    default=False,
    help="The check will be unknown if no temperature is available. Provide this option to ignore this."
)
@click.option(
    "expected_names",
    "--expect-temperature",
    multiple=True,
    default=[],
    help="Name of the temperature to expect. Can be specified multiple times. Example: board-temperature1"
)
@click.pass_context
@nagiosplugin.guarded
def system_temperature(ctx, warning_values, critical_values, use_regex, no_temperature_ok, expected_names):
    """This command reads the information from /system/health and extracts all values containing the
    word temperature in its name. Like 'board-temperature', 'board-temperature1', 'cpu-temperature', ...

    Be aware that not all devices return the same values.
    """
    check = nagiosplugin.Check()

    temperature_resource = SystemTemperatureResource(
        cmd_options=ctx.obj,
        check=check,
        warning_values=warning_values,
        critical_values=critical_values,
        use_regex=use_regex,
    )
    check.add(temperature_resource)

    check.results.add(
        nagiosplugin.Result(
            nagiosplugin.state.Ok,
            hint=f"Looks like all temperatures are OK:  {', '.join(sorted(temperature_resource.names))}"
        )
    )
    if len(temperature_resource.names) == 0 and not no_temperature_ok:
        check.results.add(
            nagiosplugin.Result(
                nagiosplugin.state.Unknown,
                hint="No temperatures found"
            )
        )

    if len(expected_names) > 0:
        missing_names = []
        for name in expected_names:
            if name not in temperature_resource.names:
                missing_names.append(name)

        if len(missing_names) > 0:
            check.results.add(
                nagiosplugin.Result(
                    nagiosplugin.state.Warn,
                    hint=f"Expected temperature(s) not found: {', '.join(missing_names)}"
                )
            )

    check.main(verbose=ctx.obj["verbose"])



# SPDX-FileCopyrightText: PhiBo DinoTools (2021)
# SPDX-License-Identifier: GPL-3.0-or-later

from typing import List, Optional

import click
import nagiosplugin

from ..cli import cli
from ..helper import logger, RouterOSVersion
from ..resource import RouterOSCheckResource


class SystemUpdateResource(RouterOSCheckResource):
    name = "Update"

    def __init__(
        self,
        cmd_options,
        check: nagiosplugin.Check,
        check_for_update: bool = False,
        latest_version: Optional[str] = None,
    ):
        super().__init__(cmd_options=cmd_options)

        self._check = check
        self._check_for_update = check_for_update
        self._installed_version = None
        self._latest_version = None
        if latest_version:
            self._latest_version = RouterOSVersion(latest_version)

    def probe(self):
        logger.info("Fetching data ...")
        if self._check_for_update:
            logger.debug("Run command to check for updates ...")
            call = self.api(
                "/system/package/update/check-for-updates"
            )
            logger.debug("Waiting that update command finished")
            # Wait until command has finished
            tuple(call)

        call = self.api.path(
            "/system/package/update"
        )
        result = tuple(call)[0]

        self._routeros_metric_values = [
            {"name": "channel", "type": None},
        ]

        installed_version = result.get("installed-version")
        if installed_version:
            self._installed_version = RouterOSVersion(installed_version)
            self._check.results.add(
                nagiosplugin.Result(
                    nagiosplugin.Ok,
                    hint=f"Installed version: {self._installed_version}"
                )
            )
        else:
            self._check.results.add(
                nagiosplugin.Result(
                    nagiosplugin.Warn,
                    hint="Unable to get installed version"
                )
            )

        latest_version = result.get("latest-version")
        if self._latest_version is None and latest_version:
            self._latest_version = RouterOSVersion(latest_version)

        if self._installed_version and self._latest_version:
            if self._installed_version < self._latest_version:
                self._check.results.add(
                    nagiosplugin.Result(
                        nagiosplugin.Critical,
                        hint=(
                            f"Update version '{self._latest_version}' available. "
                            f"Version installed '{self._installed_version}'"
                        )
                    )
                )

        status = result.get("status")
        if isinstance(status, str) and "error" in status.lower():
            self._check.results.add(
                nagiosplugin.Result(
                    nagiosplugin.Critical,
                    hint=f"Looks like there was an error: '{status}'"
                )
            )

        return self.get_routeros_metric_item(result)


class SystemUpdateChannelContext(nagiosplugin.Context):
    def __init__(self, *args, channels: Optional[List[str]] = None, **kwargs):
        super(SystemUpdateChannelContext, self).__init__(*args, **kwargs)
        self._channels = channels

    def evaluate(self, metric, resource):
        if self._channels is None or len(self._channels) == 0 or metric.value in self._channels:
            return nagiosplugin.Result(
                nagiosplugin.Ok,
                hint=f"Update channel is '{metric.value}'"
            )

        return nagiosplugin.Result(
            nagiosplugin.Warn,
            hint=f"Update channel '{metric.value}' not in list with allowed channels: {', '.join(self._channels)}"
        )


class SystemUpdateSummary(nagiosplugin.Summary):
    def ok(self, results: List[nagiosplugin.Result]):
        messages = []
        for result in results:
            messages.append(result.hint)

        if len(messages) > 0:
            return ", ".join(messages)

        return "Looks good"


@cli.command("system.update")
@click.option(
    "--channel",
    "channels",
    default=None,
    multiple=True,
    help="Allowed update channel. Repeat to use multiple values."
)
@click.option(
    "--latest-version",
    "latest_version",
    default=None,
    help=(
        "Set a version that should at least be installed. "
        "Use this if the update server is not available or if you want check with your own update policy."
    )
)
@click.option(
    "--check-for-update",
    "check_for_update",
    is_flag=True,
    default=False,
    help=(
        "Actively check for updates. "
        "This will run the command /system/package/update/check-for-updates . "
        "If you don't whant to use this feature you have to schedule a task to look for updates."
    )
)
@click.pass_context
@nagiosplugin.guarded
def system_update(ctx, channels, latest_version, check_for_update):
    check = nagiosplugin.Check()

    check.add(
        SystemUpdateResource(
            cmd_options=ctx.obj,
            check=check,
            check_for_update=check_for_update,
            latest_version=latest_version,
        ),
        SystemUpdateChannelContext(
            name="channel",
            channels=channels,
        ),
        SystemUpdateSummary(),
    )

    check.main(verbose=ctx.obj["verbose"])



# SPDX-FileCopyrightText: PhiBo DinoTools (2021)
# SPDX-License-Identifier: GPL-3.0-or-later
from datetime import date, datetime

from routeros_check.resource import RouterOSCheckResource


class TestBase:
    def test_parse_date(self):
        check = RouterOSCheckResource(cmd_options={})
        expected_date = date(year=2021, month=9, day=20)
        assert check.parse_routeros_date("sep/20/2021") == expected_date
        assert check.parse_routeros_date("2021-09-20") == expected_date

        expected_date = date(year=2022, month=10, day=13)
        assert check.parse_routeros_date("oct/13/2022") == expected_date
        assert check.parse_routeros_date("2022-10-13") == expected_date

        expected_date = date(year=2023, month=4, day=5)
        assert check.parse_routeros_date("Apr/05/2023") == expected_date
        assert check.parse_routeros_date("2023-04-05") == expected_date

        expected_date = date(year=2023, month=5, day=7)
        assert check.parse_routeros_date("May/07/2023") == expected_date
        assert check.parse_routeros_date("2023-05-07") == expected_date

        expected_date = date(year=2021, month=2, day=8)
        assert check.parse_routeros_date("Feb/08/2021") == expected_date
        assert check.parse_routeros_date("2021-02-08") == expected_date

    def test_parse_routeros_datetime(self):
        check = RouterOSCheckResource(cmd_options={})

        parsed_datetime = check.parse_routeros_datetime("sep/20/2021 13:43:12")
        assert parsed_datetime == datetime(year=2021, month=9, day=20, hour=13, minute=43, second=12)

        parsed_datetime = check.parse_routeros_datetime("oct/13/2022 22:59:59")
        assert parsed_datetime == datetime(year=2022, month=10, day=13, hour=22, minute=59, second=59)

        parsed_datetime = check.parse_routeros_datetime("Apr/05/2023 23:59:59")
        assert parsed_datetime == datetime(year=2023, month=4, day=5, hour=23, minute=59, second=59)

        parsed_datetime = check.parse_routeros_datetime("May/07/2023 22:59:59")
        assert parsed_datetime == datetime(year=2023, month=5, day=7, hour=22, minute=59, second=59)

        # Build Time 6.47.9 (long-term)
        parsed_datetime = check.parse_routeros_datetime("Feb/08/2021 12:48:33")
        assert parsed_datetime == datetime(year=2021, month=2, day=8, hour=12, minute=48, second=33)

        parsed_datetime = check.parse_routeros_datetime("2024-05-09 11:23:55")
        assert parsed_datetime == datetime(year=2024, month=5, day=9, hour=11, minute=23, second=55)

    def test_parse_routeros_time(self):
        check = RouterOSCheckResource(cmd_options={})
        assert check.parse_routeros_time_duration("-8ms15us") == -0.008015
        assert check.parse_routeros_time_duration("13m53s460ms") == 833.46
        assert check.parse_routeros_time_duration("1s") == 1
        assert check.parse_routeros_time_duration("2m1s") == 1 + 2 * 60
        assert check.parse_routeros_time_duration("3h2m1s") == 1 + 2 * 60 + 3 * 60 * 60
        assert check.parse_routeros_time_duration("4d3h2m1s") == 1 + 2 * 60 + 3 * 60 * 60 + 4 * 24 * 60 * 60
        assert check.parse_routeros_time_duration("5w4d3h2m1s") == (
            1 +
            2 * 60 +
            3 * 60 * 60 +
            4 * 24 * 60 * 60 +
            5 * 7 * 24 * 60 * 60
        )



# SPDX-FileCopyrightText: PhiBo DinoTools (2021)
# SPDX-License-Identifier: GPL-3.0-or-later
from datetime import datetime
from typing import List, Optional

import click
import nagiosplugin

from ..cli import cli
from ..helper import logger
from ..resource import RouterOSCheckResource


class SystemLicenseResource(RouterOSCheckResource):
    name = "License"

    def __init__(self, cmd_options):
        super().__init__(cmd_options=cmd_options)

        def days_left(value):
            time_delta = self.parse_routeros_datetime(value) - datetime.now()
            return int(time_delta.total_seconds()) / 60 / 60 / 24

        logger.info("Fetching information ...")
        call = self.api.path(
            "/system/resource"
        )
        result = tuple(call)[0]

        self.has_renewal = result["board-name"].lower() == "chr"

        self.deadline_datetime: Optional[datetime] = None
        self.next_renewal_datetime: Optional[datetime] = None

        self._routeros_metric_values = []

        if self.has_renewal:
            self._routeros_metric_values += [
                {"name": "level", "type": None},
                {"name": "deadline-at", "dst": "deadline-in", "type": days_left, "missing_ok": True},
                {"name": "next-renewal-at", "dst": "next-renewal-in", "type": days_left, "missing_ok": True},
            ]
        else:
            self._routeros_metric_values += [
                {"name": "nlevel", "dst": "level", "type": None},
            ]

    def probe(self):
        logger.info("Fetching data ...")
        call = self.api.path(
            "/system/license"
        )
        result = tuple(call)[0]

        if self.has_renewal:
            if "deadline-at" in result:
                self.deadline_datetime = self.parse_routeros_datetime(result["deadline-at"])
            if "next-renewal-at" in result:
                self.next_renewal_datetime = self.parse_routeros_datetime(result["next-renewal-at"])

        return self.get_routeros_metric_item(result)


class SystemLicenseRenewSummary(nagiosplugin.Summary):
    def ok(self, results: List[nagiosplugin.Result]):
        hints = []
        resource: Optional[SystemLicenseResource] = None
        for result in results:
            if result.resource:
                resource = result.resource
            if result.hint:
                hints.append(result.hint)

        if resource and resource.has_renewal:
            if resource.next_renewal_datetime:
                time_delta = resource.next_renewal_datetime - datetime.now()
                hints.append(f"Next renewal in {time_delta.days} day(s) ({resource.next_renewal_datetime})")
            if resource.deadline_datetime:
                time_delta = resource.deadline_datetime - datetime.now()
                hints.append(f"Deadline in {time_delta.days} day(s) ({resource.deadline_datetime})")

        return ", ".join(hints)


class SystemLicenseLevelContext(nagiosplugin.Context):
    def __init__(self, *args, levels=None, **kwargs):
        self._levels = levels
        super(SystemLicenseLevelContext, self).__init__(*args, **kwargs)

    def evaluate(self, metric, resource):
        if self._levels is None or len(self._levels) == 0 or metric.value in self._levels:
            return nagiosplugin.Result(
                nagiosplugin.Ok,
                hint=f"License level is '{metric.value}'"
            )

        return nagiosplugin.Result(
            nagiosplugin.Warn,
            hint=f"License level '{metric.value}' not in list with allowed levels: {', '.join(self._levels)}"
        )


@cli.command("system.license")
@click.option("--deadline-warning", default="28:", help="Number of days until deadline is reached (Default: '28:')")
@click.option("--deadline-critical", default="14:", help="Number of days until deadline is reached (Default: '14:')")
@click.option(
    "--next-renewal-warning",
    default=None,
    help="Number of days until renewal is done (Default: None, Example: '-14:')"
)
@click.option("--next-renewal-critical", default=None, help="Number of days until renewal is done (Default: None)")
@click.option(
    "--level",
    "levels",
    default=None,
    multiple=True,
    help="Allowed license levels. Repeat to use multiple values."
)
@click.pass_context
@nagiosplugin.guarded
def system_license(ctx, deadline_warning, deadline_critical, next_renewal_warning, next_renewal_critical, levels):
    resource = SystemLicenseResource(
        cmd_options=ctx.obj,
    )
    check = nagiosplugin.Check(resource)

    if resource.has_renewal:
        check.add(
            nagiosplugin.ScalarContext(
                name="deadline-in",
                warning=deadline_warning,
                critical=deadline_critical,
            ),
            nagiosplugin.ScalarContext(
                name="next-renewal-in",
                warning=next_renewal_warning,
                critical=next_renewal_critical,
            ),
            SystemLicenseRenewSummary(),
        )

    check.add(
        SystemLicenseLevelContext(
            name="level",
            levels=levels,
        )
    )

    check.main(verbose=ctx.obj["verbose"])



# SPDX-FileCopyrightText: PhiBo DinoTools (2021)
# SPDX-License-Identifier: GPL-3.0-or-later

import click
import librouteros
import librouteros.query
import nagiosplugin
from nagiosplugin.state import Ok as STATE_Ok, Warn as STATE_Warn, Critical as STATE_Critical

from ..cli import cli
from ..helper import humanize_time, logger
from ..resource import RouterOSCheckResource


class SystemUptimeResource(RouterOSCheckResource):
    name = "UPTIME"

    def __init__(self, cmd_options):
        super().__init__(cmd_options=cmd_options)

    def probe(self):
        api = self._connect_api()

        logger.info("Fetching data ...")
        call = api.path(
            "/system/resource"
        ).select(
            librouteros.query.Key("uptime"),
        )
        results = tuple(call)
        result = results[0]

        yield nagiosplugin.Metric(
            name="uptime",
            value=self.parse_routeros_time_duration(result["uptime"]),
            uom="s",
            min=0,
        )


class UptimeSimpleScalarContext(nagiosplugin.ScalarContext):
    def describe(self, metric):
        return humanize_time(metric.value)

    def evaluate(self, metric, resource):
        if str(self.critical) != "" and metric.value in self.critical:
            return self.result_cls(
                STATE_Critical,
                None,
                metric
            )

        if str(self.warning) != "" and metric.value in self.warning:
            return self.result_cls(
                STATE_Warn,
                None,
                metric
            )
        return self.result_cls(STATE_Ok, None, metric)


@cli.command("system.uptime")
@click.option(
    "--warning",
    help="State WARNING if current uptime is below this threshold",
)
@click.option(
    "--critical",
    help="State CRITICAL if current uptime is below this threshold",
    default=None,
)
@click.pass_context
@nagiosplugin.guarded
def system_uptime(ctx, warning, critical):
    """Get Uptime of a device"""
    check = nagiosplugin.Check(
        SystemUptimeResource(
            cmd_options=ctx.obj,
        ),
        UptimeSimpleScalarContext(
            name="uptime",
            warning=float(warning) if warning else None,
            critical=float(critical) if critical else None,
        )
    )

    check.main(verbose=ctx.obj["verbose"])



# SPDX-FileCopyrightText: PhiBo DinoTools (2021)
# SPDX-License-Identifier: GPL-3.0-or-later

from typing import List, Optional

import click
import librouteros
import librouteros.query
import nagiosplugin

from ..cli import cli
from ..context import ScalarPercentContext
from ..helper import logger
from ..resource import RouterOSCheckResource


class SystemDiskResource(RouterOSCheckResource):
    name = "DISK"

    def __init__(self, cmd_options):
        super().__init__(cmd_options=cmd_options)

        self.total_hdd_space: Optional[int] = None
        self._routeros_metric_values = [
            {"name": "write-sect-since-reboot", "type": int, "min": 0},
            {"name": "write-sect-total", "type": int, "min": 0},
            {"name": "bad-blocks", "type": float, "min": 0, "max": 100, "missing_ok": True},
        ]

    def probe(self):
        api = self._connect_api()

        logger.info("Fetching data ...")
        call = api.path(
            "/system/resource"
        ).select(
            librouteros.query.Key("free-hdd-space"),
            librouteros.query.Key("total-hdd-space"),
            *self.get_routeros_select_keys()
        )
        api_result_items = tuple(call)

        free_hdd_space = api_result_items[0]["free-hdd-space"]
        self.total_hdd_space = api_result_items[0]["total-hdd-space"]
        results = self.get_routeros_metric_item(api_result_items[0])

        results.append(
            nagiosplugin.Metric(
                name="free",
                value=free_hdd_space,
                uom="B",
                min=0,
                max=self.total_hdd_space,
            )
        )

        results.append(
            nagiosplugin.Metric(
                name="used",
                value=self.total_hdd_space - free_hdd_space,
                uom="B",
                min=0,
                max=self.total_hdd_space,
            )
        )

        return results


class SystemDiskSummary(nagiosplugin.summary.Summary):
    def __init__(self, result_names: List[str]):
        super().__init__()
        self._result_names = result_names

    def ok(self, results):
        msgs = []
        for result_name in self._result_names:
            msgs.append(str(results[result_name]))
        return " ".join(msgs)


@cli.command("system.disk")
@click.option(
    "--used/--free",
    is_flag=True,
    default=True,
    help="Set if used or free memory should be checked. (Default: used)",
)
@click.option(
    "--warning",
    required=True,
    help="Warning threshold in % or MB. Example (20% oder 20 = 20MB)",
)
@click.option(
    "--critical",
    required=True,
    help="Critical threshold in % or MB. Example (20% oder 20 = 20MB)",
)
@click.option(
    "--bad-blocks-warning",
    help="Warning threshold for bad blocks. Example: 20 -> 20% bad blocks",
)
@click.option(
    "--bad-blocks-critical",
    help="Critical threshold for bad blocks",
)
@click.pass_context
@nagiosplugin.guarded
def system_disk(ctx, used, warning, critical, bad_blocks_warning, bad_blocks_critical):
    check = nagiosplugin.Check(
        SystemDiskResource(
            cmd_options=ctx.obj,
        )
    )

    if used:
        check.add(nagiosplugin.ScalarContext(
            name="free",
        ))
        check.add(ScalarPercentContext(
            name="used",
            total_name="total_hdd_space",
            warning=warning,
            critical=critical
        ))
    else:
        check.add(ScalarPercentContext(
            name="free",
            total_name="total_hdd_space",
            warning=f"{warning}:",
            critical=f"{critical}:"
        ))
        check.add(nagiosplugin.ScalarContext(
            name="used",
        ))

    check.add(nagiosplugin.ScalarContext(
        name="bad-blocks",
        warning=bad_blocks_warning,
        critical=bad_blocks_critical,
    ))

    check.add(nagiosplugin.ScalarContext(
        name="write-sect-since-reboot",
    ))

    check.add(nagiosplugin.ScalarContext(
        name="write-sect-total",
    ))

    check.add(SystemDiskSummary(
        result_names=["used"] if used else ["free"]
    ))

    check.main(verbose=ctx.obj["verbose"])
